# -*- coding: UTF-8 -*-

import sys
import subprocess
import os
import commands
import ConfigParser
import threading
from settingsdnie import *
from gi import require_version
require_version('Gtk', '3.0')
from gi.repository import Gio, GLib, GObject, Gtk, Gdk
from smartcard.CardType import AnyCardType
from smartcard.CardRequest import CardRequest
from smartcard.util import toHexString
from smartcard.System import readers
from smartcard import Exceptions
from smartcard.pcsc.PCSCExceptions import EstablishContextException

GObject.threads_init()
Gdk.threads_init()



class DNIEApp:
	def __init__(self):
		self.window = Gtk.Window(Gtk.WindowType.TOPLEVEL)
		self.window.set_position(Gtk.WindowPosition.CENTER)
		self.window.set_border_width(10)
		self.window.set_size_request(660,640)
		self.window.set_title(DNIE_TITLE_WINDOW)
		self.window.set_wmclass(DNIE_TITLE_WINDOW, DNIE_TITLE_WINDOW)
		self.window.set_icon_from_file(APP_LOGO)
		self.window.connect("delete_event", self.delete_event)
		self.window.connect("destroy", self.destroy)
		self.mainbox = Gtk.VBox(False, 10)
		self.window.add(self.mainbox)

		# Cabecera de la aplicacion
		self.headbox = Gtk.HBox(False, 5)
		self.mainbox.pack_start(self.headbox, False, True, 0)
		self.logoapp = Gtk.Image()
		self.logoapp.set_from_file(APP_LOGO)
		self.logoapp.show()
		self.headbox.pack_start(self.logoapp, True, False, 0)
		self.title = Gtk.Label()
		self.title.set_markup('<span foreground="#597988" size="x-large" font-weight="ultrabold">%s</span>' % TITLE)
		self.title.set_use_markup(True)
		self.title.show()
		self.headbox.pack_start(self.title, True, False, 0)
		self.headbox.show()

		#Informacion
		self.info = Gtk.Label()
		self.info.set_markup('<span foreground="#597988" size="medium">%s</span>' % GENERAL_INFO)
		self.info.set_use_markup(True)
		self.info.show()
		self.mainbox.pack_start(self.info, False, True, 10)


		#Deteccion de dispositivo
		self.detectbox = Gtk.VBox(False, 0)
		self.mainbox.pack_start(self.detectbox, False, False, 5)
		self.detectbox.show()

		self.detectlabelnumber = Gtk.Label()
		self.detectlabelnumber.set_markup('<span foreground="#597988" size="xx-large" font-weight="ultrabold">Paso 1</span>')
		self.detectlabelnumber.set_use_markup(True)
		self.detectlabelnumber.show()
		self.detectbox.pack_start(self.detectlabelnumber, True, False, 0)
		self.detectlabel = Gtk.Label()
		self.detectlabel.set_markup('<span foreground="#597988" size="medium">%s</span>' % DETECT_INFO)
		self.detectlabel.set_use_markup(True)
		self.detectbox.pack_start(self.detectlabel, True, False, 0)
		self.detectlabel.show()

		self.detectbuttonbox = Gtk.HBox(False, 5)
		self.detectbox.pack_start(self.detectbuttonbox, False, True, 10)
		self.detectbutton = Gtk.Button.new_with_label('Detectar Dispositivo')
		self.detectbutton.connect("clicked", self.detect_smart_device)
		self.detectbutton.show()
		self.detectbuttonbox.show()
		self.detectbuttonbox.pack_start(self.detectbutton, True, False, 0)

		self.resultlabel = Gtk.Label()
		self.resultlabel.set_markup('<span font_style="italic">No se ha detectado ningún dispositivo</span>')
		self.resultlabel.set_use_markup(True)
		self.resultlabel.show()
		self.detectbox.pack_start(self.resultlabel, True, False, 10)
		self.resultfirefoxlabel = Gtk.Label()

		#Ejecutar proceso de configuracion
		self.execbox = Gtk.VBox(False, 0)
		self.mainbox.pack_start(self.execbox, False, False, 10)
		self.execbox.show()
		self.execlabelnumber = Gtk.Label()
		self.execlabelnumber.set_markup('<span foreground="#597988" size="xx-large" font-weight="ultrabold">Paso 2</span>')
		self.execlabelnumber.set_use_markup(True)
		self.execlabelnumber.show()
		self.execbox.pack_start(self.execlabelnumber, True, False, 0)

		self.execlabel = Gtk.Label()
		self.execlabel.set_markup('<span foreground="#597988" size="medium">%s<b>%s</b></span>' % (EXEC_INFO_1, EXEC_INFO_2))
		self.execlabel.set_use_markup(True)
		self.execbox.pack_start(self.execlabel, True, False, 0)
		self.execlabel.show()

		self.execbuttonbox = Gtk.HBox(False, 5)
		self.execbox.pack_start(self.execbuttonbox, False, True, 10)
		self.execbutton = Gtk.Button.new_with_label('Configurar DNIe')
		self.execbutton.connect("clicked", self.setup_reader_dnie)
		self.execbutton.set_sensitive(False)
		self.execbutton.show()
		self.execbuttonbox.show()
		self.execbuttonbox.pack_start(self.execbutton, True, False, 0)

		#Resultado del proceso
		self.spinner = Gtk.Spinner()
		self.spinner.set_size_request(25, 25)
		self.spinner_complete_image = Gtk.Image()
		self.spinner_complete_image.set_from_icon_name("process-completed-symbolic", 24)
		self.execbox.pack_start(self.spinner, True, False, 10)

		self.resultexeclabel = Gtk.Label()
		self.resultexeclabel.set_markup('')
		self.resultexeclabel.set_use_markup(True)
		self.resultexeclabel.show()
		self.execbox.pack_start(self.resultexeclabel, True, False, 10)


		self.mainbox.show()
		self.window.show()

	def detect_smart_device(self, widget, data = None):
		cardname = None
		cardatr = None
		try:
			r = readers()
			if len(r) > 0:
				cardname = r[0]
				try:
					cardtype = AnyCardType()
					cardrequest = CardRequest(timeout=1, cardType=cardtype)
					cardservice = cardrequest.waitforcard()
					cardservice.connection.connect()
					cardatr = toHexString(cardservice.connection.getATR())
					cardname = cardservice.connection.getReader()
					self.resultlabel.set_markup('<span foreground="green">Dispositivo: %s.    DNIe: insertado</span>' % cardname)
					self.execbutton.set_sensitive(True)
				except (Exceptions.CardRequestTimeoutException) as e:
					self.execbutton.set_sensitive(False)
					self.resultlabel.set_markup('<span foreground="red">Dispositivo: %s.    DNIe: no insertado</span>' % cardname)
			else:
				self.execbutton.set_sensitive(False)
				self.resultlabel.set_markup('<span foreground="red">Dispositivo: no detectado.    DNIe: no insertado</span>')
		except (EstablishContextException) as e:
			self.resultlabel.set_markup('<span foreground="red">Problemas de detección de dispositivo</span>')

		self.cardname = cardname
		self.cardatr = cardatr

	def setup_reader_dnie(self, widget, data = None):
		thread_setup = threading.Thread(target=self.setup_reader_dnie_thr)
		thread_setup.daemon = True
		thread_setup.start()

	def setup_reader_dnie_thr(self):
		if self.is_firefox_running():
			GObject.idle_add(self.resultexeclabel.set_markup, '<span foreground="red">El navegador Firefox se encuentra en ejecución</span>')
			return False
		else:
			GObject.idle_add(self.start_spinner, 0)

		if self.get_default_profile_dir() is None:
			self.create_default_profile()

		if self.has_root_ca_certificate(DNIE_ROOT_CERT_NAME):
			self.remove_root_ca_certificate(DNIE_ROOT_CERT_NAME)
		result = self.install_root_ca_certificate(DNIE_ROOT_CERT_FILE, DNIE_ROOT_CERT_NAME)
		if not result:
			GObject.idle_add(self.resultexeclabel.set_markup, '<span foreground="red">Error al insertar certificado raiz de la DGP</span>')
			GObject.idle_add(self.stop_spinner, 0)
			return result

		if self.has_root_ca_certificate(FNMT_ROOT_CERT_NAME):
			self.remove_root_ca_certificate(FNMT_ROOT_CERT_NAME)
		result = self.install_root_ca_certificate(FNMT_ROOT_CERT_FILE, FNMT_ROOT_CERT_NAME)
		if not result:
			GObject.idle_add(self.resultexeclabel.set_markup, '<span foreground="red">Error al insertar certificado raiz de la FNMT</span>')
			GObject.idle_add(self.stop_spinner, 0)
			return result
		
		if self.has_security_method(DNIE_METHOD_NAME):
			self.remove_security_method(DNIE_METHOD_NAME)
		result = self.add_security_method(DNIE_METHOD_NAME)
		if not result:
			GObject.idle_add(self.resultexeclabel.set_markup, '<span foreground="red">Error al insertar el dispositivo de seguridad</span>')
			GObject.idle_add(self.stop_spinner, 0)
			return result

		GObject.idle_add(self.resultexeclabel.set_markup, '<span foreground="green">La configuración ha sido correcta y ya puede usar su DNIe con Firefox</span>')
		GObject.idle_add(self.stop_spinner, 0)


	def add_security_method(self, name):
		pathfirefoxprofile = self.get_default_profile_dir()
		if not pathfirefoxprofile:
			return False

		cmd = [MODUTIL_CMD, '-add', "%s" % name, '-libfile', DNIE_PKCS11_LIB, '-dbdir', "%s" % pathfirefoxprofile]
		process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		process.stdin.write('\n')
		process.wait()
		return process.returncode == 0

	def has_security_method(self, name):
		pathfirefoxprofile = self.get_default_profile_dir()
		if not pathfirefoxprofile:
			return False

		cmd = '%s -list -dbdir "%s"' % (MODUTIL_CMD, pathfirefoxprofile)
		status, output = commands.getstatusoutput(cmd)
		return name in output

	def remove_security_method(self, name):
		pathfirefoxprofile = self.get_default_profile_dir()
		if not pathfirefoxprofile:
			return False

		cmd = [MODUTIL_CMD, '-delete', "%s" % name, '-dbdir', "%s" % pathfirefoxprofile]
		process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		process.stdin.write('\n')
		process.wait()
		return process.returncode == 0


	def install_root_ca_certificate(self, certificate_file, ca_name):
		pathfirefoxprofile = self.get_default_profile_dir()
		cmd = '%s -A -d "%s" -i %s -n "%s" -t "TCu,TCu,TCu"'
		cmd = cmd % (CERTUTIL_CMD, pathfirefoxprofile, certificate_file, ca_name)
		status, output = commands.getstatusoutput(cmd)
		return status == 0

	def has_root_ca_certificate(self, ca_name):
		pathfirefoxprofile = self.get_default_profile_dir()
		cmd = '%s -L -n "%s" -d "%s"'
		cmd = cmd % (CERTUTIL_CMD, ca_name, pathfirefoxprofile)
		status, output = commands.getstatusoutput(cmd)
		return status == 0

	def remove_root_ca_certificate(self, ca_name):
		pathfirefoxprofile = self.get_default_profile_dir()
		if not pathfirefoxprofile:
			return False

		cmd = '%s -D -d "%s" -n "%s"'
		cmd = cmd % (CERTUTIL_CMD, pathfirefoxprofile, ca_name)
		status, output = commands.getstatusoutput(cmd)
		return status == 0

	def is_firefox_running(self):
		cmd = 'ps -e | grep firefox | wc -l'
		status, output = commands.getstatusoutput(cmd)
		return int(output) != 0

	def get_default_profile_dir(self):
		user_dir = os.path.expanduser('~')
		ff_dir = os.path.join(user_dir, '.mozilla', 'firefox')
		if not os.path.exists(ff_dir):
			return

		config = ConfigParser.ConfigParser()
		config.readfp(file(os.path.join(ff_dir, 'profiles.ini')))
		profiles = [section for section in config.sections() if config.has_option(section, 'Name')]

		if len(profiles) == 1:
			default = profiles[0]
		else:
			for profile in profiles:
				if (config.has_option(profile, 'Default') and config.get(profile, 'Default') == '1'):
					default = profile
		path = config.get(default, 'Path')
		return os.path.join(ff_dir, path)

	def create_default_profile(self):
		cmd = '%s -CreateProfile "default"' % FIREFOX_CMD
		status, output = commands.getstatusoutput(cmd)
		return status == 0

	def start_spinner(self, data):
		self.spinner.start()
		self.resultexeclabel.hide()
		self.spinner.show()
		self.execbutton.set_sensitive(False)

	def stop_spinner(self, data):
		self.spinner.stop()
		self.spinner.hide()
		self.resultexeclabel.show()

	def delete_event(self, widget, data = None):
		Gtk.main_quit()

	def destroy(self, widget, data = None):
		Gtk.main_quit()

def main():
	Gtk.main()

if __name__=="__main__":
	app = DNIEApp()
	main()
